package cpu

// State contains information about which registers are currently in use.
type State uint64

// Free will reset the reserved and used status which means the register can be allocated again.
func (s *State) Free(reg Register) {
	*s &= ^(1 << reg)
}

// IsUsed returns true if the register is currently in use and holds a value.
func (s *State) IsUsed(reg Register) bool {
	return *s&(1<<reg) != 0
}

// Use marks a register to be currently in use which means it must be preserved across function calls.
func (s *State) Use(reg Register) {
	*s |= (1 << reg)
}

// FindFree tries to find a free register in the given slice of registers.
func (s *State) FindFree(registers []Register) (Register, bool) {
	for _, reg := range registers {
		if !s.IsUsed(reg) {
			return reg, true
		}
	}

	return 0, false
}

// MustFindFree tries to find a free register and panics if it could not be found.
func (s *State) MustFindFree(registers []Register) Register {
	register, exists := s.FindFree(registers)

	if !exists {
		panic("no free registers")
	}

	return register
}
