package scope

import (
	"git.urbach.dev/cli/q/src/ast"
	"git.urbach.dev/cli/q/src/cpu"
	"git.urbach.dev/cli/q/src/eval"
	"git.urbach.dev/cli/q/src/token"
)

// Stack is a stack of scopes.
type Stack struct {
	Scopes []*Scope
}

// AddVariable adds a new variable to the current scope.
func (stack *Stack) AddVariable(variable *Variable) {
	stack.CurrentScope().AddVariable(variable)
}

// CurrentScope returns the current scope.
func (stack *Stack) CurrentScope() *Scope {
	return stack.Scopes[len(stack.Scopes)-1]
}

// PopScope removes the scope at the top of the stack.
func (stack *Stack) PopScope() {
	stack.Scopes = stack.Scopes[:len(stack.Scopes)-1]
}

// PushScope pushes a new scope to the top of the stack.
func (stack *Stack) PushScope(body ast.AST, buffer []byte) *Scope {
	s := &Scope{}

	if len(stack.Scopes) > 0 {
		lastScope := stack.Scopes[len(stack.Scopes)-1]
		s.State = lastScope.State
		s.Variables = make([]*Variable, 0, len(lastScope.Variables))
		s.InLoop = lastScope.InLoop

		for _, v := range lastScope.Variables {
			count := ast.Count(body, buffer, token.Identifier, v.Name)

			if count == 0 {
				continue
			}

			s.Variables = append(s.Variables, &Variable{
				Name: v.Name,
				Value: eval.Register{
					Typ:      v.Value.Typ,
					Alive:    count,
					Register: v.Value.Register,
				},
			})
		}
	}

	stack.Scopes = append(stack.Scopes, s)
	return s
}

// UseVariable reduces the lifetime of the variable in all scopes.
func (stack *Stack) UseVariable(variable *Variable) {
	for _, scope := range stack.Scopes {
		if scope.InLoop {
			continue
		}

		local := scope.VariableByName(variable.Name)

		if local == nil {
			continue
		}

		local.Value.Use()

		if !local.Value.IsAlive() {
			scope.Free(local.Value.Register)
		}
	}
}

// VariableByName returns the variable with the given name or `nil` if it doesn't exist.
func (stack *Stack) VariableByName(name string) *Variable {
	return stack.CurrentScope().VariableByName(name)
}

// VariableByRegister returns the variable that occupies the given register or `nil` if none occupy the register.
func (stack *Stack) VariableByRegister(register cpu.Register) *Variable {
	for _, v := range stack.CurrentScope().Variables {
		if v.Value.Register == register {
			return v
		}
	}

	return nil
}
